from numba import jit, njit, cfunc, prange
import numpy as np
from astropy import constants as c

from scipy import interpolate

C_KMS = c.c.to('km/s').value


@cfunc("float64[:](float64[:], float64, float64)")
def normal(x, mu, sigma):
    """
    A normalized Gaussian distribution with amplitude such that the area under the curve integrates to 1.
    :param x: np.array
        x positions to evaluate the function at.
    :param mu: float
        the mean value
    :param sigma: float
        the standard deviation
    :return: np.array
        f(x)
    """
    return 1 / (sigma * np.sqrt(2 * np.pi)) * np.exp(-(x - mu) ** 2 / (2 * sigma ** 2))


@cfunc("float64[:](float64[:], float64, float64, float64)")
def gaussian(x, A, mu, sigma):
    """
    A Gaussian distribution with an arbitrary amplitude.
    :param x: np.array
        x positions to evaluate the function at.
    :param A: float
        The amplitude.
    :param mu: float
        the mean value
    :param sigma: float
        the standard deviation
    :return: np.array
        f(x)
    """
    return A * np.exp(-(x - mu) ** 2 / (2 * sigma ** 2))


@cfunc("float64(float64, int64)")
def truncate(number, digits) -> float:
    stepper = np.power(10, digits)
    return int(stepper * number) / stepper


@njit
def find_closest(x, val):
    """Finds the index and corresponding value in x which is closest to some val.

    Args:
        x (np.ndarray): The array to search.
        val (float): The value to be closest to.

    Returns:
        int: The index of the closest member
        float: The value at that index.
    """
    diffs = np.abs(x - val)
    loc = np.nanargmin(diffs)
    return loc, x[loc]


@cfunc("float64[:](float64[:], float64)")
def cosmological_redshift(wave, z):
    """

    :param wave:
    :param z: float
        The dimensionless redshift parameter of the object where 1+z = λ_obs/λ_emit
    :return: float
        The redshift-corrected wavelength
    """
    return wave / (1 + z)


@cfunc("float64(float64)")
def calc_redshift_sqrt(vel):
    """
    Calculate the redshift from the velocity.

    :param vel: float
        The velocity of the object in km/s.
    :return wave: np.ndarray
        The redshift-corrected wavelength array.
    """
    v = vel / C_KMS
    return np.sqrt((1+v)/(1-v)) - 1


@cfunc("float64(float64)")
def calc_redshift_exp(vel):
    """
    Calculate the the redshift from the velocity.

    :param vel: float
        The velocity of the object in km/s.
    :return wave: np.ndarray
        The redshift-corrected wavelength array.
    """
    v = vel / C_KMS
    return np.exp(v) - 1


@njit
def correct_extinction(wave, flux, ebv, r_v):
    """
    Deredden a flux vector using the CCM 1989 parameterization
    Returns an array of the unreddened flux

    :param wave: np.ndarray
        The wavelength vector.
    :param flux: np.ndarray
        The flux vector to be corrected.
    :param ebv: float
        E(B-V) in magintudes.
    :param r_v: optional, float
        specifies the ratio of total selective
        extinction R(V) = A(V)/E(B-V). If not specified,
        then r_v = 3.1
    :return flux: np.ndarray
        The unreddened calibrated flux array, same number of
        elements as wave

    NOTES:
    0. This function was taken from BADASS3, created by Remington Sexton, https://github.com/remingtonsexton/BADASS3.
       All notes below come from the original BADASS documentation.

    1. (From BADASS:) This function was converted from the IDL Astrolib procedure
       last updated in April 1998. All notes from that function
       (provided below) are relevant to this function

    2. (From IDL:) The CCM curve shows good agreement with the Savage & Mathis (1979)
       ultraviolet curve shortward of 1400 A, but is probably
       preferable between 1200 and 1400 A.
    3. (From IDL:) Many sightlines with peculiar ultraviolet interstellar extinction
       can be represented with a CCM curve, if the proper value of
       R(V) is supplied.
    4. (From IDL:) Curve is extrapolated between 912 and 1000 A as suggested by
       Longo et al. (1989, ApJ, 339,474)
    5. (From IDL:) Use the 4 parameter calling sequence if you wish to save the
       original flux vector.
    6. (From IDL:) Valencic et al. (2004, ApJ, 616, 912) revise the ultraviolet CCM
       curve (3.3 -- 8.0 um-1).	But since their revised curve does
       not connect smoothly with longer and shorter wavelengths, it is
       not included here.

    7. For the optical/NIR transformation, the coefficients from
       O'Donnell (1994) are used

    # >>> ccm_unred([1000, 2000, 3000], [1, 1, 1], 2 )
    array([9.7976e+012, 1.12064e+07, 32287.1])
    """
    assert wave.size == flux.size, "Wave and flux must have the same size!"

    x = 10000.0 / wave
    npts = wave.size
    a = np.zeros(npts)
    b = np.zeros(npts)

    ###############################
    # Infrared

    good = np.where((x > 0.3) & (x < 1.1))[0]
    a[good] = 0.574 * x[good] ** 1.61
    b[good] = -0.527 * x[good] ** 1.61

    ###############################
    # Optical & Near IR

    good = np.where((x >= 1.1) & (x < 3.3))[0]
    y = x[good] - 1.82

    c1 = np.array([1.0, 0.104, -0.609, 0.701, 1.137, -1.718, -0.827, 1.647, -0.505])
    c2 = np.array([0.0, 1.952, 2.908, -3.989, -7.985, 11.102, 5.491, -10.805, 3.347])

    order = len(c1)
    poly_a = np.zeros(len(y))
    for i in prange(order):
        poly_a += c1[::-1][i] * y ** (order-1-i)
    a[good] = poly_a

    order = len(c2)
    poly_b = np.zeros(len(y))
    for j in prange(order):
        poly_b += c2[::-1][j] * y ** (order-1-j)
    b[good] = poly_b
    # a[good] = np.polyval(c1[::-1], y)
    # b[good] = np.polyval(c2[::-1], y)

    ###############################
    # Mid-UV

    good = np.where((x >= 3.3) & (x < 8))[0]
    y = x[good]
    F_a = np.zeros(good.size)
    F_b = np.zeros(good.size)
    good1 = np.where(y > 5.9)[0]

    if good1.size > 0:
        y1 = y[good1] - 5.9
        F_a[good1] = -0.04473 * y1 ** 2 - 0.009779 * y1 ** 3
        F_b[good1] = 0.2130 * y1 ** 2 + 0.1207 * y1 ** 3

    a[good] = 1.752 - 0.316 * y - (0.104 / ((y - 4.67) ** 2 + 0.341)) + F_a
    b[good] = -3.090 + 1.825 * y + (1.206 / ((y - 4.62) ** 2 + 0.263)) + F_b

    ###############################
    # Far-UV

    good = np.where((x >= 8) & (x <= 11))[0]
    y = x[good] - 8.0
    c1 = [-1.073, -0.628, 0.137, -0.070]
    c2 = [13.670, 4.257, -0.420, 0.374]

    order = len(c1)
    poly_a = np.zeros(len(y))
    for i in prange(order):
        poly_a += c1[::-1][i] * y ** (order-1-i)
    a[good] = poly_a

    order = len(c2)
    poly_b = np.zeros(len(y))
    for j in prange(order):
        poly_b += c2[::-1][j] * y ** (order-1-j)
    b[good] = poly_b
    # a[good] = np.polyval(c1[::-1], y)
    # b[good] = np.polyval(c2[::-1], y)

    # Applying Extinction Correction

    a_v = r_v * ebv
    a_lambda = a_v * (a + b / r_v)

    flux = flux * 10.0 ** (0.4 * a_lambda)

    return flux  # ,a_lambda


@njit
def make_bins(wavs):
    """
    NOTES:
        This function has been taken and adapted from the SpectRes package, https://github.com/ACCarnall/SpectRes
        to improve performance.

    ORIGINAL DOCSTRING:

    Given a series of wavelength points, find the edges and widths
    of corresponding wavelength bins.
    """
    edges = np.zeros(wavs.shape[0]+1)
    widths = np.zeros(wavs.shape[0])
    edges[0] = wavs[0] - (wavs[1] - wavs[0])/2
    widths[-1] = (wavs[-1] - wavs[-2])
    edges[-1] = wavs[-1] + (wavs[-1] - wavs[-2])/2
    edges[1:-1] = (wavs[1:] + wavs[:-1])/2
    widths[:-1] = edges[1:-1] - edges[:-2]

    return edges, widths


@njit
def spectres(new_wavs, spec_wavs, spec_fluxes, spec_errs, fill=None, verbose=True):

    """
    NOTES:
        This function has been taken and adapted from the SpectRes package, https://github.com/ACCarnall/SpectRes
        to improve performance.  Please note that unlike the original package, spec_errs is now a REQUIRED argument.

    ORIGINAL DOCSTRING:

    Function for resampling spectra (and optionally associated
    uncertainties) onto a new wavelength basis.

    Parameters
    ----------

    new_wavs : numpy.ndarray
        Array containing the new wavelength sampling desired for the
        spectrum or spectra.

    spec_wavs : numpy.ndarray
        1D array containing the current wavelength sampling of the
        spectrum or spectra.

    spec_fluxes : numpy.ndarray
        Array containing spectral fluxes at the wavelengths specified in
        spec_wavs, last dimension must correspond to the shape of
        spec_wavs. Extra dimensions before this may be used to include
        multiple spectra.

    spec_errs : numpy.ndarray (optional)
        Array of the same shape as spec_fluxes containing uncertainties
        associated with each spectral flux value.

    fill : float (optional)
        Where new_wavs extends outside the wavelength range in spec_wavs
        this value will be used as a filler in new_fluxes and new_errs.

    verbose : bool (optional)
        Setting verbose to False will suppress the default warning about
        new_wavs extending outside spec_wavs and "fill" being used.

    Returns
    -------

    new_fluxes : numpy.ndarray
        Array of resampled flux values, first dimension is the same
        length as new_wavs, other dimensions are the same as
        spec_fluxes.

    new_errs : numpy.ndarray
        Array of uncertainties associated with fluxes in new_fluxes.
        Only returned if spec_errs was specified.
    """

    # Rename the input variables for clarity within the function.
    old_wavs = spec_wavs
    old_fluxes = spec_fluxes
    old_errs = spec_errs

    # Make arrays of edge positions and widths for the old and new bins

    old_edges, old_widths = make_bins(old_wavs)
    new_edges, new_widths = make_bins(new_wavs)

    # Generate output arrays to be populated
    new_fluxes = np.zeros(np.shape(old_fluxes[..., 0]) + new_wavs.shape)

    if old_errs.shape != old_fluxes.shape:
        raise ValueError("Spec_errs must be the same shape "
                         "as spec_fluxes.")
    else:
        new_errs = np.copy(new_fluxes)

    start = 0
    stop = 0
    warned = False

    # Calculate new flux and uncertainty values, looping over new bins
    for j in range(new_wavs.shape[0]):

        # Add filler values if new_wavs extends outside of spec_wavs
        if (new_edges[j] < old_edges[0]) or (new_edges[j+1] > old_edges[-1]):
            new_fluxes[..., j] = fill

            new_errs[..., j] = fill

            if (j == 0 or j == new_wavs.shape[0]-1) and verbose and not warned:
                warned = True
                print("\nSpectres: new_wavs contains values outside the range "
                      "in spec_wavs, new_fluxes and new_errs will be filled "
                      "with the value set in the 'fill' keyword argument. \n")
            continue

        # Find first old bin which is partially covered by the new bin
        while old_edges[start+1] <= new_edges[j]:
            start += 1

        # Find last old bin which is partially covered by the new bin
        while old_edges[stop+1] < new_edges[j+1]:
            stop += 1

        # If new bin is fully inside an old bin start and stop are equal
        if stop == start:
            new_fluxes[..., j] = old_fluxes[..., start]
            if old_errs is not None:
                new_errs[..., j] = old_errs[..., start]

        # Otherwise multiply the first and last old bin widths by P_ij
        else:
            start_factor = ((old_edges[start+1] - new_edges[j])
                            / (old_edges[start+1] - old_edges[start]))

            end_factor = ((new_edges[j+1] - old_edges[stop])
                          / (old_edges[stop+1] - old_edges[stop]))

            old_widths[start] *= start_factor
            old_widths[stop] *= end_factor

            # Populate new_fluxes spectrum and uncertainty arrays
            f_widths = old_widths[start:stop+1]*old_fluxes[..., start:stop+1]
            new_fluxes[..., j] = np.sum(f_widths, axis=-1)
            new_fluxes[..., j] /= np.sum(old_widths[start:stop+1])

            e_wid = old_widths[start:stop+1]*old_errs[..., start:stop+1]

            new_errs[..., j] = np.sqrt(np.sum(e_wid**2, axis=-1))
            new_errs[..., j] /= np.sum(old_widths[start:stop+1])

            # Put back the old bin widths to their initial values
            old_widths[start] /= start_factor
            old_widths[stop] /= end_factor

    # If errors were supplied return both new_fluxes and new_errs.
    return new_fluxes, new_errs



from multiprocessing.sharedctypes import Value
from textwrap import fill
import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import toml
import os
import numexpr as ne
import plotly.subplots as psp
import plotly.graph_objects as pgo

from scipy.stats import reciprocal
from skopt.searchcv import BayesSearchCV
from skopt.plots import plot_objective, plot_histogram
from skopt.space import Real, Integer, Categorical

import bifrost.maths as maths
import bifrost


class NeuralNet:

    # A class for creating and training neural networks to detect optical coronal lines

    __slots__ = ['min_wave', 'max_wave', 'spec_size', 'n_extra_params', 'loss', 'optimizer', 'metrics', 'model', 
                 'regressor', 'search_model']
    config_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "config", "neural.net.toml"))
    config = toml.load(config_path)

    def __init__(self, spec_size=None, min_wave=None, max_wave=None, loss=tf.keras.losses.BinaryCrossentropy,
                 optimizer=tf.keras.optimizers.Adam, metrics=('binary_accuracy',)):
        """
        Initialize the neural network.

        :param spec_size: int
            Number of pixels that the logarithmically-spaced wavelength grid for the neural network should take up,
            defaults to 101.
        :param min_wave: float
            Minimum wavelength of the grid, defaults to line-50.
        :param max_wave: float
            Maximum wavelength of the grid, defaults to line+50.
        :param loss: tensorflow.keras.losses
            Loss function for the neural network. Defaults to BinaryCrossentropy.
        :param optimizer: tensorflow.keras.optimizers
            Optimizer for the neural network.  Defaults to Adam.
        :param metrics: iterable
            Metrics for the neural network.  Defaults to ('binary_accuracy',).
        :return self:
            The NeuralNet instance.
        """

        param_distribs = {
            # Minimum of 3 dense layers (not counting the input and activation layers)
            "n_layers": Integer(3, 6, prior="uniform"),
            "n_neurons": Integer(0, 100, prior="uniform"),
            "learning_rate": Real(1e-10, 1e-1, prior="log-uniform")
        }

        self.regressor = tf.keras.wrappers.scikit_learn.KerasRegressor(self._build_model)
        self.search_model = BayesSearchCV(self.regressor, param_distribs, n_iter=50, cv=3, n_jobs=3, refit=True, verbose=9)
        self.model = None

        # Create the overall wavelength grid and the keras model
        self.min_wave = min_wave
        self.max_wave = max_wave
        self.spec_size = spec_size
        # Number of extra parameters to add to input data aside from flux:
        # 1. RMS of the flux
        self.n_extra_params = 0
        # Model hyperparameters
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics
    
    def _build_model(self, n_layers, n_neurons, learning_rate):
        """
        Helper function to build a keras model with the given parameters

        :param n_layers: int
            Number of dense layers of the neural network
        :param n_neurons: int
            Number of neurons per dense layer of the neural network
        :param learning_rate: float
            The learning rate of the neural network
        :return model: tensorflow.keras.models.Sequential
            The keras neural network.
        """
        model = tf.keras.models.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(self.spec_size+self.n_extra_params,)))
        for layer in range(n_layers):
            model.add(tf.keras.layers.Dense(n_neurons, activation="relu"))
        model.add(tf.keras.layers.Dense(1))
        model.compile(loss=self.loss(from_logits=True), optimizer=self.optimizer(learning_rate=learning_rate), metrics=self.metrics)
        return model
    
    def _get_wave_limits(self, line, target_line=0, size=100_000):
        training_parameters = NeuralNet.config["training_parameters"]

        # Get shape for outputs based on wavelength
        if type(line) in (str, np.str_):
            if line not in training_parameters["line_sets"].keys():
                wavelength = [training_parameters[line]['wavelength']]
                line = [line]
                SHAPE = (size,1)
            else:
                line = [line]
                wavelength = [training_parameters[line[0]]['wavelength']]
                for li in training_parameters["line_sets"][line[0]]:
                    line.append(li)
                    wavelength.append(training_parameters[li]['wavelength'])
                SHAPE = (size,len(line))
        elif type(line) in (list, np.ndarray):
            for li in line:
                if li in training_parameters["line_sets"]:
                    for lli in training_parameters["line_sets"][li]:
                        if lli not in line:
                            line.append(lli)
            wavelength = [training_parameters[li]['wavelength'] for li in line]
            SHAPE = (size,len(line))
        else:
            raise ValueError("line must be one of: int, float, list, np.ndarray")
        line = np.asarray(line)
        wavelength = np.asarray(wavelength)

        if self.min_wave is None:
            self.min_wave = np.min(wavelength) - 50
        if self.max_wave is None:
            self.max_wave = np.max(wavelength) + 50
        if self.spec_size is None:
            self.spec_size = int(np.max(wavelength) - np.min(wavelength) + 101)
        
        return line, wavelength, SHAPE


    def train(self, line="generic_line", target_line=0, size=100_000, epochs=11, out_path="neuralnet_training_data", save=True,
              save_path="bifrost.neuralnet.h5", plot=True):
        """
        The main function for training the neural network to recognize a specific optical coronal line ("line") using simulated data

        :param line: str, iterable
            The line name(s) to include in the simulated training data
        :param target_line: int
            The index of which line in the "line" argument is the line to train the neural network to detect. Default = 0.
        :param size: int
            The size of the simulated dataset. Default = 100_000.
        :param epochs: int
            The number of epochs to train the data for. Default = 11.
        :param out_path: str
            The output path for saving the neural network h5 file, or any plots. Default = 'neuralnet_training_data'
        :param save: bool
            Whether or not to save the trained network as an h5 file. Default = True.
        :param save_path: str
            The name of the h5 file to be saved if save=True. Default = 'bifrost.neuralnet.h5'
        :param plot: bool
            Whether or not to plot a subset of the simulated training data (100). Default = True.
        :return None:
        """

        training_parameters = NeuralNet.config["training_parameters"]
        # Make RNG seeds
        power = int(np.log10(size))
        rng_seeds = np.random.randint(int(10**power), int(10**(power+1)-1), size)

        line, wavelength, SHAPE = self._get_wave_limits(line, target_line, size)

        # Initialize output parameters, baselines, and noise
        output_parameters = {}
        baselines = np.ones(size)
        constrained = []               

        # Initialize parameters that are constant for every line: random noise and the power slope
        for param in ("noise", "power_slope"):
            if training_parameters[param]['dist'] == 'uniform':
                p1, p2 = 'min', 'max'
            elif training_parameters[param]['dist'] in ('normal', 'lognormal'):
                p1, p2 = 'mean', 'std'
            else:
                raise ValueError(f"Unrecognized dist config parameter '{training_parameters[param]['dist']}'")
            output_parameters[param] = eval(f"np.random.{training_parameters[param]['dist']}({training_parameters[param][p1]}," 
                        f"{training_parameters[param][p2]}, {size})")

        # Initialize parameters that are different for every line
        for param in ("amp", "fwhm", "voff", "h3", "h4", "eta"):
            output_parameters[param] = np.zeros(SHAPE)
            # Create the distributions
            output_parameters[param] = np.zeros(SHAPE)
            for i, li in enumerate(line):
                # Check for profile-specific features
                if param in ("h3", "h4") and training_parameters[li]["profile"] not in ("GH", "random"):
                    output_parameters[param][:, i] = np.nan
                    continue
                if param == "eta" and training_parameters[li]["profile"] not in ("V", "random"):
                    output_parameters[param][:, i] = np.nan
                    continue

                # Generate keywords
                if training_parameters[li][param]['dist'] == 'uniform':
                    p1, p2 = 'min', 'max'
                elif training_parameters[li][param]['dist'] in ('normal', 'lognormal'):
                    p1, p2 = 'mean', 'std'
                elif training_parameters[li][param]['dist'] == 'constrained':
                    constrained.append((i, param))
                    continue
                else:
                    raise ValueError(f"Unrecognized dist config parameter '{training_parameters[li][param]['dist']}'")

                output_parameters[param][:, i] = eval(f"np.random.{training_parameters[li][param]['dist']}"
                    f"({training_parameters[li][param][p1]}, {training_parameters[li][param][p2]}, {size})")
        
        # Make some ratio of the simulated spectra have no line of interest
        # output_parameters["amp"][:output_parameters["amp"].shape[0]//2, target_line] = 0.
        for i in range(len(line)):
            if type(training_parameters[line[i]]["detect_ratio"]) is float:
                # Small adjustment to the ratio to make the end result be more close to the input ratio 
                # (accounting for those with SNR < the threshold)
                epsilon = 1.2 if training_parameters[line[i]]["detect_ratio"] < 0.833 else 1

                idx = int((1 - epsilon*training_parameters[line[i]]["detect_ratio"]) * len(output_parameters["amp"][:, i]))
                ids = np.random.choice(range(len(output_parameters["amp"][:, i])), idx, replace=False)
                output_parameters["amp"][ids, i] = 0.
        # Need to make 2 passes in case the line that doublets depend on appears after it does in the line list
        for i in range(len(line)):
            if type(training_parameters[line[i]]["detect_ratio"]) is str: 
                li = np.where(line == training_parameters[line[i]]["detect_ratio"])[0][0]
                noline = np.where(output_parameters["amp"][:, li] == 0.)[0]
                output_parameters["amp"][noline, i] = 0.
        # Special case for [Fe X] -- remove any detections where [O I] is not detected
        if "FeX_6374" in line and "OI_6302" in line and "OI_6365" in line:
            oi1 = np.where(line == "OI_6302")[0][0]
            fi = np.where(line == "FeX_6374")[0][0]
            noline = np.where(output_parameters["amp"][:, oi1] == 0.)[0]
            output_parameters["amp"][noline, fi] = 0.
            # yesline = np.where(output_parameters["amp"][:, oi1] > 0.)[0]
            # idy = int(training_parameters[line[fi]]["detect_ratio"] * len(yesline))
            # output_parameters["amp"][yesline[:idy], fi] = 0.

        for c in constrained:
            constraint = training_parameters[line if type(line) in (str, np.str_) else line[c[0]]][c[1]]
            assert constraint['dist'] == 'constrained'
            l_expr, n_expr = constraint['constraint'][0], constraint['constraint'][1]

            # Aliasing for nice-looking numexpr strings in the toml file:
            c_l = 0 if type(line) in (str, np.str_) else np.where(line == l_expr)[0][0]
            amp = output_parameters["amp"][:, c_l]
            noise = output_parameters["noise"][c_l]
            fwhm = output_parameters["fwhm"][:, c_l]
            voff = output_parameters["voff"][:, c_l]
            h3 = output_parameters["h3"][:, c_l]
            h4 = output_parameters["h4"][:, c_l]
            eta = output_parameters["eta"][:, c_l]
            # Randomized coefficients for use in numexpr
            randc = np.random.uniform(0.7, 1, len(amp))
            randc2 = np.random.uniform(0.3, 0.5, len(amp))
        
            output_parameters[c[1]][:, c[0]] = ne.evaluate(n_expr)

        # Parameter constraints
        # Ensure certain parameters remain positive
        # output_parameters["amp"][output_parameters["amp"] < 0] = 0
        if output_parameters["eta"] is not None:
            output_parameters["eta"][output_parameters["eta"] < 0] = 0
            output_parameters["eta"][output_parameters["eta"] > 1] = 1
        # Ensure certain parameters remain nonzero
        output_parameters["fwhm"][output_parameters["fwhm"] <= 0] = 1
        output_parameters["noise"][output_parameters["noise"] <= 0] = 0.01


        profile = [training_parameters[li]["profile"] for li in line]
        h_moments = np.dstack((output_parameters["h3"], output_parameters["h4"]))

        print("Preparing data for neural network training...")
        # breakpoint()
        # Create simulated data
        stack = bifrost.Stack.quick_sim_stack(
            wavelength, baselines, output_parameters["amp"], output_parameters["fwhm"],
            output_parameters["voff"], output_parameters["power_slope"], h_moments=h_moments, eta_mixes=output_parameters["eta"],
            noise_amplitudes=output_parameters["noise"], profiles=profile, min_wave=self.min_wave,
            max_wave=self.max_wave, size=self.spec_size, seeds=rng_seeds, out_path=out_path, progress_bar=True)

        labels = np.where(np.abs(output_parameters["amp"][:, target_line] / output_parameters["noise"]) > NeuralNet.config["SNR"], 1, 0)
        # labels = np.where(output_parameters["amp"][:, target_line] > 0, 1, 0)
        # labels = np.concatenate(([0] * (size // 2), [1] * (size // 2)))
        # labels = np.array([(0,1)[stack[i].data["snr"][target_line] >= NeuralNet.config["SNR"]] for i in range(len(stack))])
        # For debugging purposes:
        if plot:
            stack.plot_spectra(out_path, backend='pyplot', spectra=list(stack.keys())[0:100], 
                            _range=(self.min_wave, self.max_wave), normalized=True,
                            ylim=(-5,5), title_text={label: f"Line: {labels[i]}, SNR: {output_parameters['amp'][i, target_line]/output_parameters['noise'][i]:.2f}, "
                            f"Amp: {output_parameters['amp'][i, target_line]:.2f}, Noise: {output_parameters['noise'][i]:.2f}" for i, label in enumerate(stack)})

        # Split data into train, validation, and test sets
        ind = np.array(range(len(stack)))
        np.random.shuffle(ind)
        i1 = len(stack)*8//10
        i2 = len(stack)*9//10
        train_set = ind[0:i1]
        valid_set = ind[i1:i2]
        test_set = ind[i2:]

        # Resample each spectrum onto full_wave_grid, preserving flux and error
        train_data, tmu, tsig = self.normalize(np.array([stack[i].flux for i in train_set], dtype=float))
        # train_err = np.array([np.nanstd(stack[i].flux) for i in train_set], dtype=float).reshape((len(train_data), 1))
        # train_data = np.hstack((train_data, train_err))
        train_labels = np.array([labels[i] for i in train_set], dtype=int)

        valid_data, vmu, vsig = self.normalize(np.array([stack[i].flux for i in valid_set], dtype=float))
        # valid_err = np.array([np.nanstd(stack[i].flux) for i in valid_set], dtype=float).reshape((len(valid_data), 1))
        # valid_data = np.hstack((valid_data, valid_err))
        valid_labels = np.array([labels[i] for i in valid_set], dtype=int)

        test_data, ttmu, ttsig = self.normalize(np.array([stack[i].flux for i in test_set], dtype=float))
        # test_err = np.array([np.nanstd(stack[i].flux) for i in test_set], dtype=float).reshape((len(test_data), 1))
        # test_data = np.hstack((test_data, test_err))
        test_labels = np.array([labels[i] for i in test_set], dtype=int)

        self.search_model.fit(train_data, train_labels, epochs=epochs, validation_data=(valid_data, valid_labels))

        self.model = self.search_model.best_estimator_.model
        self.model.evaluate(test_data, test_labels, verbose=2)
        if save:
            self.model.save(os.path.join(out_path, save_path))
        if plot:
            plot_objective(self.search_model.optimizer_results_[0], dimensions=["learning_rate", "n_layers", "n_neurons"],
                           n_minimum_search=int(1e8))
            plt.savefig(os.path.join(out_path, "modelsearch.cornerplot.pdf"), dpi=300, bbox_inches="tight")
            plt.close()
            for i in range(3):
                plot_histogram(self.search_model.optimizer_results_[0], i)
                plt.savefig(os.path.join(out_path, f"modelsearch.histogram{i}.pdf"), dpi=300, bbox_inches="tight")
                plt.close()

    @staticmethod
    def normalize(data):
        """
        A helper function to normalize training data to have a mean of 0 and standard deviation of 1.

        :param data: np.ndarray
            The array of data to be normalized.
        :return: np.ndarray
            The normalized data.
        """
        mu = np.nanmean(data, axis=-1)
        sigma = np.nanstd(data, axis=-1)
        return ((data.T - mu) / sigma).T, mu, sigma

    def load(self, path, line, target_line=0):
        """
        Load an already-trained keras model into the neural network via an h5 file.

        :param path: str
            The path to the h5 file to load.
        :return None:
        """
        self._get_wave_limits(line, target_line, size=100_000)
        self.model = tf.keras.models.load_model(path)

    def predict(self, test_stack, p_layer="sigmoid", plot=False, out_path=None):
        """
        Use the trained neural network to predict the presence of coronal lines in real data.

        :param test_stack: bifrost.Stack
            A Stack object containing the spectra to test for coronal lines.
        :param p_layer: str
            The type of activation layer to use for the prediction results. Default = 'sigmoid'.
        :param plot: bool
            Whether or not to plot the resultant spectra with their predictions. Default = False.
        :param out_path: str
            The output path to save plots if plot = True.
        :return predictions: np.ndarray
            Array of prediction values (confidences between 0-1) for each spectrum in the input stack.
        """
        # Construct wavelength grid for the inputs
        wave_grid = np.geomspace(self.min_wave, self.max_wave, self.spec_size)
        # Correct for redshift and galactic extinction before testing
        test_stack.correct_spectra()
        # Resample input spectra onto the proper shaped wavelength grid for input to the neural network
        test_data, ttmu, ttsig = self.normalize(np.array([
            maths.spectres(wave_grid, test_stack[i].wave, test_stack[i].flux, test_stack[i].error, fill=np.nan)[0]
            for i in range(len(test_stack))], dtype=float))

        # test_data, ttmu, ttsig = self.normalize(np.array([test_stack[i].flux for i in range(len(test_stack))], dtype=float))
        # test_err = ((np.array([test_stack[i].error for i in range(len(test_stack))], dtype=float).T - ttmu) / ttsig).T

        # test_err = np.array([np.nanstd(test_stack[i].flux) for i in range(len(test_stack))], dtype=float).reshape((len(test_data), 1))
        # test_data = np.hstack((test_data, test_err))

        # Make sure there are no nans or infs so that the neural net still works
        for i in range(test_data.shape[0]):
            if np.isfinite(np.nanmedian(test_data[i, :])):
                test_data[i, ~np.isfinite(test_data[i, :])] = np.nanmedian(test_data[i, :])
            else:
                test_data[i, :] = 0.

        probability_model = tf.keras.Sequential([self.model, tf.keras.layers.Activation(p_layer)])
        predictions = probability_model.predict(test_data)
        print(predictions)

        if plot:
            if out_path is None:
                out_path = "neuralnet_training_data"
            test_stack.plot_spectra(out_path, backend='pyplot',
                _range=(self.min_wave, self.max_wave), ylim=(-5,5),
                spectra=np.asarray(list(test_stack.keys()))[np.where(predictions > 0.99)[0]],  # only plot the really confident spectra
                title_text={label: f"NN Confidence: {predictions[k]}" for k, label in enumerate(test_stack.keys())},
                normalized=True)

        return predictions
    
    def convolve(self, wave, flux, error, p_layer="sigmoid", plot=True, out_path=None):
        # Take the neural network as a sliding window along the length of the spectrum,
        # generating a confidence as a function of wavelength

        # Gather the left and right wavelength bounds for each window
        slide_size = len(wave)-self.spec_size+1
        left_wave = np.geomspace(wave[0], wave[-self.spec_size], slide_size)
        right_wave = np.geomspace(wave[self.spec_size-1], wave[-1], slide_size)

        cwave = np.full(slide_size, fill_value=np.nan)
        data = np.full((slide_size, self.spec_size), fill_value=np.nan)

        # Go through and resample the flux for each window onto the proper wavelength grid
        for i in range(slide_size):
            wgi = np.geomspace(left_wave[i], right_wave[i], self.spec_size)
            cwave[i] = np.nanmedian(wgi)
            data[i, :], _, _ = self.normalize(maths.spectres(wgi, wave, flux, error, fill=np.nan)[0])
            if np.isfinite(np.nanmedian(data[i, :])):
                data[i, ~np.isfinite(data[i, :])] = np.nanmedian(data[i, :])
            else:
                data[i, :] = 0.
        
        # Perform the neural network predictions across each window
        probability_model = tf.keras.Sequential([self.model, tf.keras.layers.Activation(p_layer)])
        conf = probability_model.predict(data).T[0]
        cflux, _, _ = self.normalize(maths.spectres(cwave, wave, flux, error, fill=np.nan)[0])

        # Plot the probability vs. wavelength
        if plot:
            if out_path is None:
                out_path = "neuralnet_training_data/nn.convolve.confidence.html"
            # fig, ax = plt.subplots(figsize=(10,5))
            # ax.plot(cwave, conf, 'k-')
            # ax.set_xlabel(r'$\lambda (\AA)$')
            # ax.set_ylabel(r'Line Confidence')
            # plt.savefig(os.path.join(out_path, "nn.convolve.confidence.pdf"), dpi=300, bbox_inches='tight')
            # plt.close()
            fig = psp.make_subplots(rows=1, cols=1)
            fig.add_trace(pgo.Scatter(x=cwave, y=conf, line=dict(color='red', width=2), name='Confidence', showlegend=False))
            fig.add_trace(pgo.Scatter(x=cwave, y=cflux, line=dict(color='black', width=1), name='Flux', showlegend=False))
            fig.update_layout(
                xaxis_title=r'$\lambda (A)$',
                yaxis_title=r'Confidence | Flux (normalized)',
                template='plotly_white'
            )
            fig.write_html(out_path, include_mathjax="cdn")

        return cwave, conf